/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#include "tcp/tcpsegmentfactory.h"

#include <logfunc.hpp>

#include "boost_assert_handler.h"
#include "tcp/tcpheader.h"
#include "tcp/tcpipv4segment.h"
#include "tcp/tcpipv6segment.h"

using namespace std;
using boost::asio::ip::address;
using boost::asio::ip::address_v4;
using boost::asio::ip::address_v6;
using boost::asio::ip::tcp_raw_protocol;
using I2n::Logger::GlobalLogger;

//-----------------------------------------------------------------------------
// TcpSegmentFactory
//-----------------------------------------------------------------------------

/**
 * @brief Creates a TCP segment from the input stream @c std::istream.
 *
 * @param protocol The segment's network layer protocol, IPv4 or IPv6.
 * @param is The input stream.
 *
 * @return A TCP Segment object.
 */
TcpSegmentItem TcpSegmentFactory::create_tcp_segment(
        const tcp_raw_protocol::socket::protocol_type &protocol,
        std::istream &is
)
{
    BOOST_ASSERT( (tcp_raw_protocol::v4() == protocol) || (tcp_raw_protocol::v6() == protocol) );

    TcpSegmentItem tcp_segment;

    if ( tcp_raw_protocol::v4() == protocol )
    {
        tcp_segment.reset( new TcpIpv4Segment() );
    }
    else if ( tcp_raw_protocol::v6() == protocol )
    {
        tcp_segment.reset( new TcpIpv6Segment() );
    }
    else
    {
        BOOST_ASSERT( !"Invalid TCP Segment Type." );                                                                                   //lint !e506
    }

    if ( !tcp_segment->read( is ) )
    {
        GlobalLogger.error() << "Could not read TCP Segment." << endl;
    }

    return tcp_segment;
}

/**
 * @brief Creates a TCP SYN request segment.
 *
 * @param protocol The segment's network layer protocol, IPv4 or IPv6.
 * @param source_address The local address originating the ping.
 * @param destination_address The remote address where to ping.
 * @param source_port The local TCP port.
 * @param destination_port The remote TCP port.
 * @param sequence_number The segment's sequence number to aid in matching RST
 * or SYN/ACK replies to this SYN request. May be zero.
 *
 * @return A TCP SYN request segment object.
 */
TcpSegmentItem TcpSegmentFactory::create_tcp_segment_syn_request(
        const tcp_raw_protocol::socket::protocol_type &protocol,
        const address &source_address,
        const address &destination_address,
        const uint16_t source_port,
        const uint16_t destination_port,
        const uint16_t sequence_number
)
{
    BOOST_ASSERT( (tcp_raw_protocol::v4() == protocol) || (tcp_raw_protocol::v6() == protocol) );

    TcpSegmentItem tcp_segment;

    if ( tcp_raw_protocol::v4() == protocol )
    {
        tcp_segment = create_tcp_ipv4_segment_syn_request(
                source_address, destination_address,
                source_port, destination_port, sequence_number );
    }
    else if ( tcp_raw_protocol::v6() == protocol )
    {
        tcp_segment = create_tcp_ipv6_segment_syn_request(
                source_address, destination_address,
                source_port, destination_port, sequence_number );
    }
    else
    {
        BOOST_ASSERT( !"Invalid TCP Segment Type." );                                                                           //lint !e506
    }

    return tcp_segment;
}

/**
 * @brief Creates a TCP SYN request segment over IPv4.
 *
 * @param source_address The local address originating the ping.
 * @param destination_address The remote address where to ping.
 * @param source_port The local TCP port.
 * @param destination_port The remote TCP port.
 * @param sequence_number The segment's sequence number to aid in matching RST
 * or SYN/ACK replies to this SYN request. May be zero.
 *
 * @return A TCP SYN request segment object.
 */
TcpSegmentItem TcpSegmentFactory::create_tcp_ipv4_segment_syn_request(
        const address &source_address,
        const address &destination_address,
        const uint16_t source_port,
        const uint16_t destination_port,
        const uint16_t sequence_number
)
{
    BOOST_ASSERT( source_address.is_v4() );
    BOOST_ASSERT( destination_address.is_v4() );

    // (5 words of 32 bits = 5 * 4 bytes = 20 bytes)
    const uint8_t header_size_in_words = 5; // size in units of 32 bits
    const uint16_t window_size_in_octets = 32768;

    // Create a TCP header for a SYN request.
    TcpHeader tcp_header;
    tcp_header.set_source_port( source_port ); // assign a random ephemeral port number
    tcp_header.set_destination_port( destination_port );
    tcp_header.set_sequence_number( sequence_number );
    tcp_header.set_header_length( header_size_in_words );
    tcp_header.set_synchronize( true );
    tcp_header.set_window_size( window_size_in_octets ); // window size

    // Calculate the checksum of the TCP header
    address_v4::bytes_type src_address = source_address.to_v4().to_bytes();
    address_v4::bytes_type dst_address = destination_address.to_v4().to_bytes();
    uint16_t cksum = tcp_header.calculate_tcp_checksum(
            src_address, dst_address, NULL, 0
    );
    tcp_header.set_checksum( cksum );

    TcpSegmentItem tcp_segment( new TcpIpv4Segment( tcp_header ) );

    return tcp_segment;
}

/**
 * @brief Creates a TCP SYN request segment over IPv6.
 *
 * @param source_address The local address originating the ping.
 * @param destination_address The remote address where to ping.
 * @param source_port The local TCP port.
 * @param destination_port The remote TCP port.
 * @param sequence_number The segment's sequence number to aid in matching RST
 * or SYN/ACK replies to this SYN request. May be zero.
 *
 * @return A TCP SYN request segment object.
 */
TcpSegmentItem TcpSegmentFactory::create_tcp_ipv6_segment_syn_request(
        const address &source_address,
        const address &destination_address,
        const uint16_t source_port,
        const uint16_t destination_port,
        const uint16_t sequence_number
)
{
    BOOST_ASSERT( source_address.is_v6() );
    BOOST_ASSERT( destination_address.is_v6() );

    // (5 words of 32 bits = 5 * 4 bytes = 20 bytes)
    const uint8_t header_size_in_words = 5; // size in units of 32 bits
    const uint16_t window_size_in_octets = 32768;

    // Create a TCP header for a SYN request.
    TcpHeader tcp_header;
    tcp_header.set_source_port( source_port ); // assign a random ephemeral port number
    tcp_header.set_destination_port( destination_port );
    tcp_header.set_sequence_number( sequence_number );
    tcp_header.set_header_length( header_size_in_words );
    tcp_header.set_synchronize( true );
    tcp_header.set_window_size( window_size_in_octets ); // window size

    // Calculate the checksum
    address_v6::bytes_type src_address = source_address.to_v6().to_bytes();
    address_v6::bytes_type dst_address = destination_address.to_v6().to_bytes();
    uint16_t cksum = tcp_header.calculate_tcp_checksum(
            src_address, dst_address, NULL, 0
    );
    tcp_header.set_checksum( cksum );

    TcpSegmentItem tcp_segment( new TcpIpv6Segment( tcp_header ) );

    return tcp_segment;
}
